import java.awt.*;
import javax.swing.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.Random;
import java.util.LinkedList;

public class MatrixCodez implements Runnable {
    // lowercase so don't clash
    static final int screen_width  = 800;
    static final int screen_height = 450;
    static String fetch;

    public static void main (String ... args) {
        fetch = (args.length > 0) ? args[0] : "http://forum.java.sun.com/rss/rssthreads.jsp?forumID=31";

        SwingUtilities.invokeLater(new MatrixCodez());
    }

    interface Channel {
        int nextChar();
        void reset();
    }

    public void run () {
        final JFrame frame = new JFrame("Matrix Codez");

        final BufferedImage image = new BufferedImage(screen_width, screen_height, BufferedImage.TYPE_INT_ARGB );

        // start off black-ish

        Graphics2D g2d = image.createGraphics();

        final Color       letterGreen = new Color(126, 255, 126);
        final Color       greyWash = new Color(0, 0, 0, 12);

        g2d.setPaint(Color.BLACK);
        g2d.fillRect(0, 0, screen_width, screen_height);

        g2d.setPaint(new Color(126, 255, 126, 32));
        g2d.fillRect(0, 0, screen_width, screen_height);

        final JComponent panel = new JComponent() {
            public void paint (Graphics g) {
                g.drawImage(image, 0, 0, screen_width, screen_height, null);
            }
        };

        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.getContentPane().setLayout(new BorderLayout());
        frame.getContentPane().add(panel, BorderLayout.CENTER);
        frame.setVisible(true);
        frame.setSize(screen_width, screen_height);

        final Random random = new Random();
        final int cell_width = 12;
        final int cell_height = 16;
        final int columns = (screen_width + cell_width - 1) / cell_width;
        final int rows = (screen_height + cell_height - 1) / cell_height;

        class Chars {
            LinkedList<Channel> channels = new LinkedList<Channel>();
            long time;

            public Channel nextChannel () {
                if (channels.isEmpty()) reset();

                Channel channel = channels.get(0);

                channels.remove(0);

                return channel;
            }

            {
                items = "Loading...#Please wait.".split("#");

                reset();
            }

            String[] items;

            void loadItems () {
                try {
                    System.out.println("fetching " + fetch);

                    URL url = new URL(fetch);

                    URLConnection connection = url.openConnection();

                    connection.connect();

                    InputStreamReader in = new InputStreamReader(connection.getInputStream());

                    // read all data into a blob
                    StringBuilder buffer = new StringBuilder();
                    char buf[] = new char[1024];

                    for (int len; (len = in.read(buf)) != -1; )
                        buffer.append(buf, 0, len);

                    synchronized (this) {
                        // not the way to handle XML really, but...
                        items = buffer.toString().split("<item>");
                    }
                } catch (Exception ex) {
                    ex.printStackTrace(System.out);
                }
            }


            void reset() {
                try {
                    if (time == 0 || (time + 60000 < System.currentTimeMillis())) {
                        time = System.currentTimeMillis();
                        new Thread(new Runnable() { public void run () { loadItems(); } }).start();
                    }

                    String[] items;

                    synchronized (this) {
                        items = this.items;
                    }

                    for (String item : items) {
                        final StringReader reader = new StringReader(item);

                        channels.add(new Channel() {
                            public void reset() { }
                            public int nextChar () {
                                try {
                                    return reader.read();
                                } catch (Exception ex) {
                                    throw new RuntimeException (ex);
                                }
                            }
                        });
                    }

                    //in = FileInputStream("/home/pete/Library/Shakespeare/The Tempest.html");
                } catch (Exception ex) {
                    throw new RuntimeException (ex);
                }
            }
        }

        final Chars chars = new Chars();

        // some character streams
        class Stream {
            int x;
            int y;
            Channel chan;

            Stream () {
                reset();
            }

            void reset () {
                if (chan == null)
                    chan = chars.nextChannel();

                x = random.nextInt(columns);
                y = random.nextInt(rows * 3) - rows;
            }

            void update (Graphics2D g2d, FontMetrics fm) {
                ++y;

                if (y > 0) {
                    do {
                        int ch = chan.nextChar();

                        if (ch == -1) {
                            chan = null;
                            reset();
                            return;
                        }

                        String text = String.valueOf((char)ch);

                        int text_width = fm.stringWidth(text);

                        g2d.drawString(text, x * cell_width + (cell_width - text_width) / 2, y * cell_height);
                    } while (random.nextInt(7) == 0);

                    if (y > rows || random.nextInt(15) == 0)
                        reset();
                }
            }
        }

        final Stream[] streams = new Stream[(2*columns)/3];

        for (int i = 0; i < streams.length; ++i)
            streams[i] = new Stream();

        final Font font = new Font("Inconsolata", Font.BOLD, 14);

        new Timer(25, new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                Graphics2D  g2d = image.createGraphics();

                g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
                g2d.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
                g2d.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);

                g2d.setColor(greyWash);
                g2d.fillRect(0, 0, screen_width, screen_height);

                g2d.setColor(letterGreen);
                g2d.setFont(font);

                FontMetrics fm = g2d.getFontMetrics();

                for (Stream stream : streams)
                    stream.update(g2d, fm);

                panel.repaint();
            }
        }).start();
    }
}
